
def factory(func):
	"""
		Factory decorator. Executes the decorated
		method/function and returns 'self' at the end.
	"""
	def inner(self, *args, **kw):
		func(self, *args, **kw)
		return self
	return inner



class Argument(object):
	def __init__(self, *args, **kw):
		super(Argument, self).__init__()
		self.args = args # positional arugments
		self.kw = kw # keyword arguments


class FileArgument(Argument):
	def __init__(self, *args, **kw):
		super(FileArgument, self).__init__(*args, **kw)

	@classmethod
	def mode(cls, file_mode, encoding=None):
		def wrapper(*args, **kw):
			obj = cls(*args, **kw)
			obj.kw["type"] = argparse.FileType(file_mode, encoding=encoding)
			return obj
		return wrapper



from abc import ABC

from cvargparse.utils import factory
from cvargparse.argument import Argument as Arg


class BaseFactory(ABC):
	'''

	'''
	def __init__(self, initial=None):
		super(BaseFactory, self).__init__()
		self.args = initial or []

	@factory
	def add(self, *args, **kwargs):
		self.args.append(Arg(*args, **kwargs))


	def get(self):
		return self.args


class ArgFactory(BaseFactory):
	'''

	'''
	@factory
	def batch_size(self):
		self.add('--batch_size', '-b', type=int, default=32, help='batch size')


	@factory
	def epochs(self):
		self.add('--epochs', '-e', type=int, default=30, help='number of epochs')


	@factory
	def debug(self):
		self.add('--debug', action='store_true', help='enable chainer debug mode')


	@factory
	def seed(self):
		self.add('--seed', type=int, default=None, help='random seed')


	@factory
	def weight_decay(self, default=5e-3):
		self.add('--decay', type=float, default=default, help='weight decay')


	@factory
	def learning_rate(self, lr=1e-2, lrs=10, lrd=1e-1, lrt=1e-6):
		self.add('--learning_rate', '-lr', type=float, default=lr, help='learning rate')
		self.add('--lr_shift', '-lrs', type=int, default=lrs, help='learning rate shift interval (in epochs)')
		self.add('--lr_decrease_rate', '-lrd', type=float, default=lrd, help='learning rate decrease')
		self.add('--lr_target', '-lrt', type=float, default=lrt, help='learning rate target')



__version__ = "0.2.0"

from cvargparse.argument import Argument
from cvargparse.argument import FileArgument
Arg = Argument

from cvargparse.factory import ArgFactory
from cvargparse.factory import BaseFactory
from cvargparse.parser import BaseParser
from cvargparse.parser import GPUParser

__all__ = [
	"Arg",
	"Argument",
	"FileArgument",
	"ArgFactory",
	"BaseFactory",
	"BaseParser",
	"GPUParser",
]



import argparse
import logging

from cvargparse.argument import Argument as Arg
from cvargparse.factory import BaseFactory

class BaseParser(argparse.ArgumentParser):

	def __init__(self, arglist=[], nologging=False, sysargs=None, *args, **kw):
		self.__nologging = nologging
		self.__sysargs = sysargs
		self._groups = {}

		super(BaseParser, self).__init__(*args, **kw)

		self.add_args(arglist)

		if not self.has_logging: return

		group = self.add_argument_group("Logger arguments")
		group.add_argument(
			'--logfile', type=str, default='',
			help='file for logging output')

		group.add_argument(
			'--loglevel', type=str, default='INFO',
			help='logging level. see logging module for more information')

		self.__args = None


	def get_group(self, name):
		return self._groups.get(name)

	def has_group(self, name):
		return name in self._groups


	def add_argument_group(self, title, *args, **kwargs):
		group = super(BaseParser, self).add_argument_group(title=title, *args, **kwargs)
		self._groups[title] = group
		return group


	def add_args(self, arglist, group_name=None, group_kwargs={}):

		if isinstance(arglist, BaseFactory):
			arglist = arglist.get()

		if group_name is None:
			group = self
		elif self.has_group(group_name):
			group = self.get_group(group_name)
		else:
			group = self.add_argument_group(group_name, **group_kwargs)

		for arg in arglist:
			if isinstance(arg, Arg):
				group.add_argument(*arg.args, **arg.kw)
			else:
				group.add_argument(*arg[0], **arg[1])

	@property
	def args(self):
		if self.__args is None:
			self.__args = self.parse_args(self.__sysargs)

		return self.__args


	@property
	def has_logging(self):
		return not self.__nologging

	def init_logger(self, simple=False):
		if not self.has_logging: return
		fmt = '%(message)s' if simple else '%(levelname)s - [%(asctime)s] %(filename)s:%(lineno)d [%(funcName)s]: %(message)s'
		logging.basicConfig(
			format=fmt,
			level=getattr(logging, self.args.loglevel.upper(), logging.DEBUG),
			filename=self.args.logfile or None,
			filemode="w")


class GPUParser(BaseParser):
	def __init__(self, *args, **kw):
		super(GPUParser, self).__init__(*args, **kw)
		self.add_argument(
			"--gpu", "-g", type=int, nargs="+", default=[-1],
			help="which GPU to use. select -1 for CPU only")



from enum import Enum
from enum import EnumMeta

from cvargparse import Arg

class MetaBaseType(EnumMeta):
	"""
		MetaType for the base enumeration type.
	"""

	def __contains__(cls, item):
		"""
			Redefines the "in" operation.
		"""
		if isinstance(item, str):
			return item.lower() in cls.as_choices()
		else:
			return super(MetaBaseType, cls).__contains__(item)

	def __getitem__(cls, key):
		"""
			Redefines the "[]" operation.
		"""
		return cls.as_choices()[key.lower()]

class BaseChoiceType(Enum, metaclass=MetaBaseType):
	"""
		Enum base type. Can be used to define argument choices.
		It also enables to quickly creat, get and display the defined choices.
	"""

	@classmethod
	def as_choices(cls):
		return {e.name.lower(): e for e in cls}

	@classmethod
	def get(cls, key):
		if isinstance(key, str):
			return cls[key] if key in cls else cls.Default
		elif isinstance(key, cls):
			return key
		else:
			raise ValueError("Unknown optimizer type: \"{}\"".format(key.__class__.__name__))

	@classmethod
	def as_arg(cls, name, short_name=None, default=None, help_text=None, **kwargs):

		args = ["--{}".format(name)]

		if short_name is not None:
			args.append("-{}".format(short_name))

		if help_text is None:
			help_text = "choices for \"{}\"".format(name)

		default = default or cls.Default.name.lower()

		help_text += " (default: {})".format(default)

		return Arg(*args,
			type=str, default=default,
			choices=cls.as_choices(),
			help=help_text,
			**kwargs)
